/*
 * Copyright (C) 2005  Network Applied Communication Laboratory Co., Ltd.
 *
 * This file is part of Rast.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA
 */

#include <apr_strings.h>
#include <ruby.h>
#include <st.h>

#undef PACKAGE_NAME
#undef PACKAGE_TARNAME
#undef PACKAGE_VERSION
#undef PACKAGE_STRING
#include "rast/config.h"
#include "rast/text_index.h"
#include "rast/db.h"
#include "rast/query.h"

static VALUE eError, eRastError, eAprError, eBDBError;
static VALUE cPool, cTerm, cTermFrequency, cCandidate, cQueryResult;
static VALUE cResult, cResultTerm, cResultItem;
static VALUE cProperty, cTextIndexer, cEncoding;
static VALUE mVNUM;

static void
raise_error(rast_error_t *error)
{
    VALUE error_class;
    char buf[RAST_BUFSIZ];

    switch (error->type) {
    case RAST_ERROR_TYPE_RAST:
        error_class = eRastError;
        break;
    case RAST_ERROR_TYPE_APR:
        error_class = eAprError;
        break;
    case RAST_ERROR_TYPE_BDB:
        error_class = eBDBError;
        break;
    default:
        error_class = eError;
    };

#ifdef RAST_DEBUG
    snprintf(buf, RAST_BUFSIZ, "%s:%d: %s",
             error->file, error->line, error->message);
#else
    strncpy(buf, error->message, RAST_BUFSIZ);
#endif
    buf[RAST_BUFSIZ - 1] = '\0';
    rast_error_destroy(error);
    rb_raise(error_class, "%s", buf);
}

static void
pool_free(apr_pool_t *pool)
{
    if (pool) {
        apr_pool_destroy(pool);
    }
}

static VALUE
pool_alloc(VALUE klass)
{
    return Data_Wrap_Struct(klass, NULL, pool_free, NULL);
}

static int
raise_on_pool_failure(int retcode)
{
    rb_memerror();
    return -1; /* prevent compiler warnings */
}

static VALUE
pool_initialize(VALUE self)
{
    apr_status_t status;
    apr_pool_t *pool;

    status = apr_pool_create_ex(&pool, NULL, raise_on_pool_failure, NULL);
    if (status != APR_SUCCESS) {
        rb_memerror();
    }
    DATA_PTR(self) = pool;
    return Qnil;
}

static VALUE
pool_new()
{
    VALUE obj;

    obj = pool_alloc(cPool);
    pool_initialize(obj);
    return obj;
}

static apr_pool_t *
get_pool(VALUE self)
{
    if (TYPE(self) != T_DATA ||
        RDATA(self)->dfree != (RUBY_DATA_FUNC) pool_free) {
        rb_raise(rb_eTypeError,
                 "wrong argument type %s (expected Rast::Pool)",
                 rb_obj_classname(self));
    }
    return (apr_pool_t *) DATA_PTR(self);
}
 
static rast_encoding_module_t *
get_encoding_module(const char *name)
{
    rast_encoding_module_t *encoding_module;
    rast_error_t *error;

    error = rast_get_encoding_module(name, &encoding_module);
    if (error != RAST_OK) {
        raise_error(error);
    }
    return encoding_module;
}

static VALUE
encoding_new(char *name)
{
    rast_error_t *error;
    rast_encoding_module_t *encoding_module;

    error = rast_get_encoding_module(name, &encoding_module);
    if (error != RAST_OK) {
        rast_error_destroy(error);
        return Qnil;
    }
    return Data_Wrap_Struct(cEncoding, NULL, NULL, encoding_module);
}

static VALUE
encoding_register_tokenize(VALUE self, VALUE text)
{
    VALUE vpool;
    apr_pool_t *pool;
    rast_encoding_module_t *encoding_module;
    rast_tokenizer_t *tokenizer;
    rast_token_t token;
    VALUE ary;
    
    StringValue(text);
    vpool = pool_new();
    pool = get_pool(vpool);

    Data_Get_Struct(self, rast_encoding_module_t, encoding_module);
    tokenizer = rast_register_tokenizer_create(pool, encoding_module,
                                               RSTRING(text)->ptr,
                                               RSTRING(text)->len);
    ary = rb_ary_new2(3);
    while (!rast_register_tokenizer_is_done(tokenizer)) {
        rast_register_tokenizer_get_current(tokenizer, &token);
        rb_ary_store(ary, 0, rb_str_new(token.ptr, token.nbytes));
        rb_ary_store(ary, 1, INT2NUM(token.pos));
        rb_ary_store(ary, 2, token.is_complete ? Qtrue : Qfalse);
        rb_yield(ary);
        rast_register_tokenizer_next(tokenizer);
    }
    return Qnil;
}

static VALUE
encoding_search_tokenize(VALUE self, VALUE text)
{
    VALUE vpool;
    apr_pool_t *pool;
    rast_encoding_module_t *encoding_module;
    rast_tokenizer_t *tokenizer;
    rast_token_t token;
    VALUE ary;
    
    StringValue(text);
    vpool = pool_new();
    pool = get_pool(vpool);

    Data_Get_Struct(self, rast_encoding_module_t, encoding_module);
    tokenizer = rast_search_tokenizer_create(pool, encoding_module,
                                               RSTRING(text)->ptr,
                                               RSTRING(text)->len);
    ary = rb_ary_new2(3);
    while (!rast_search_tokenizer_is_done(tokenizer)) {
        rast_search_tokenizer_get_current(tokenizer, &token);
        rb_ary_store(ary, 0, rb_str_new(token.ptr, token.nbytes));
        rb_ary_store(ary, 1, INT2NUM(token.pos));
        rb_ary_store(ary, 2, token.is_complete ? Qtrue : Qfalse);
        rb_yield(ary);
        rast_search_tokenizer_next(tokenizer);
    }
    return Qnil;
}

static VALUE
encoding_normalize_text(VALUE self, VALUE src_text)
{
    VALUE vpool;
    apr_pool_t *pool;
    rast_encoding_module_t *encoding_module;
    char *dst_text;
    rast_size_t dst_len;
    
    StringValue(src_text);
    vpool = pool_new();
    pool = get_pool(vpool);
    Data_Get_Struct(self, rast_encoding_module_t, encoding_module);
    encoding_module->normalize_text(pool,
                                    RSTRING(src_text)->ptr,
                                    RSTRING(src_text)->len,
                                    &dst_text, &dst_len);
    return rb_str_new(dst_text, dst_len);
}

static VALUE
encoding_normalize_chars(VALUE self, VALUE src)
{
    VALUE vpool;
    apr_pool_t *pool;
    rast_encoding_module_t *encoding_module;
    char *dst;
    rast_size_t dst_len;
    
    StringValue(src);
    vpool = pool_new();
    pool = get_pool(vpool);
    Data_Get_Struct(self, rast_encoding_module_t, encoding_module);
    encoding_module->normalize_chars(pool,
                                     RSTRING(src)->ptr, RSTRING(src)->len,
                                     &dst, &dst_len);
    return rb_str_new(dst, dst_len);
}

typedef struct {
    rast_text_indexer_t *indexer;
    apr_pool_t *pool;
} text_indexer_data_t;

static void
text_indexer_free(text_indexer_data_t *data)
{
    if (data) {
        apr_pool_destroy(data->pool);
        xfree(data);
    }
}

static VALUE
text_indexer_new(rast_text_indexer_t *indexer)
{
    return Data_Wrap_Struct(cTextIndexer, NULL, text_indexer_free, indexer);
}

static VALUE
text_indexer_add(VALUE self, VALUE str)
{
    text_indexer_data_t *data;
    rast_error_t *error;

    StringValue(str);
    Data_Get_Struct(self, text_indexer_data_t, data);
    error = rast_text_indexer_add(data->indexer,
                                  RSTRING(str)->ptr, RSTRING(str)->len, NULL);
    if (error != RAST_OK) {
        raise_error(error);
    }
    return Qnil;
}

static VALUE
text_indexer_commit(VALUE self)
{
    text_indexer_data_t *data;
    rast_error_t *error;

    Data_Get_Struct(self, text_indexer_data_t, data);
    error = rast_text_indexer_commit(data->indexer);
    if (error != RAST_OK) {
        raise_error(error);
    }
    return Qnil;
}

typedef struct {
    rast_text_index_t *index;
    apr_pool_t *pool;
    int closed;
} text_index_data_t;

static void
text_index_free(text_index_data_t *data)
{
    if (data) {
        if (!data->closed) {
            rast_text_index_close(data->index);
        }
        apr_pool_destroy(data->pool);
        xfree(data);
    }
}

static VALUE
text_index_alloc(VALUE klass)
{
    return Data_Wrap_Struct(klass, NULL, text_index_free, NULL);
}

static VALUE
text_index_initialize(int argc, VALUE *argv, VALUE self)
{
    VALUE name, vflags, vblock_size;
    int flags = RAST_RDWR;
    rast_encoding_module_t *encoding_module;
    rast_error_t *error;
    apr_pool_t *pool;
    rast_text_index_t *index;
    text_index_data_t *data;
    rast_size_t block_size;

    rb_scan_args(argc, argv, "12", &name, &vflags, &vblock_size);
    if (!NIL_P(vflags)) {
        flags = NUM2INT(vflags);
    }
    if (NIL_P(vblock_size)) {
        block_size = 512;
    }
    else {
        block_size = NUM2INT(vblock_size);
    }
    apr_pool_create_ex(&pool, NULL, raise_on_pool_failure, NULL);
    encoding_module = get_encoding_module("utf8");
    error = rast_text_index_open(&index, StringValuePtr(name), flags,
                                 encoding_module, NULL, 0, block_size, pool);
    if (error != RAST_OK) {
        raise_error(error);
    }
    data = ALLOC(text_index_data_t);
    data->index = index;
    data->pool = pool;
    data->closed = 0;
    DATA_PTR(self) = data;
    return Qnil;
}

static VALUE
text_index_close(VALUE self)
{
    text_index_data_t *data;
    rast_error_t *error;

    Data_Get_Struct(self, text_index_data_t, data);
    error = rast_text_index_close(data->index);
    if (error != RAST_OK) {
        raise_error(error);
    }
    data->closed = 1;
    return Qnil;
}

static VALUE
text_index_sync(VALUE self)
{
    text_index_data_t *data;
    rast_error_t *error;

    Data_Get_Struct(self, text_index_data_t, data);
    error = rast_text_index_sync(data->index);
    if (error != RAST_OK) {
        raise_error(error);
    }
    return Qnil;
}

static VALUE
text_index_register(int argc, VALUE *argv, VALUE self)
{
    text_index_data_t *data;
    rast_error_t *error;
    rast_text_indexer_t *indexer;
    rast_doc_id_t doc_id;
    apr_pool_t *pool;

    Data_Get_Struct(self, text_index_data_t, data);
    if (argc < 1) {
        rb_raise(rb_eArgError, "wrong number of arguments (0 for 1)");
    }
    doc_id = NUM2INT(argv[0]);
    apr_pool_create(&pool, NULL);
    error = rast_text_index_register(data->index, doc_id, &indexer, pool);
    if (error != RAST_OK) {
        raise_error(error);
    }
    if (argc == 1) {
        return text_indexer_new(indexer);
    }
    else {
        int i;

        for (i = 1; i < argc; i++) {
            StringValue(argv[i]);
            error = rast_text_indexer_add(indexer,
                                          RSTRING(argv[i])->ptr,
                                          RSTRING(argv[i])->len, NULL);
            if (error != RAST_OK) {
                raise_error(error);
            }
        }
        error = rast_text_indexer_commit(indexer);
        if (error != RAST_OK) {
            raise_error(error);
        }
        return Qnil;
    }
}

static VALUE
query_result_new(rast_query_result_t *result)
{
    rast_term_t *t;
    rast_candidate_t *c;
    rast_term_frequency_t *tf;
    VALUE terms, term, candidates, term_frequencies, term_frequency, candidate;

    terms = rb_ary_new();
    APR_RING_FOREACH(t, &result->terms, rast_term_t, link) {
        term = rb_funcall(cTerm, rb_intern("new"), 2,
                          rb_tainted_str_new2(t->term), INT2NUM(t->doc_count));
        rb_ary_push(terms, term);
    }
    candidates = rb_ary_new();
    APR_RING_FOREACH(c, &result->candidates, rast_candidate_t, link) {
        term_frequencies = rb_ary_new();
        APR_RING_FOREACH(tf, &c->terms, rast_term_frequency_t, link) {
            term_frequency = rb_funcall(cTermFrequency, rb_intern("new"), 2,
                                        INT2NUM(tf->count), INT2NUM(tf->pos));
            rb_ary_push(term_frequencies, term_frequency);
        }
        candidate = rb_funcall(cCandidate, rb_intern("new"), 2,
                               INT2NUM(c->doc_id), term_frequencies);
        rb_ary_push(candidates, candidate);
    }
    return rb_funcall(cQueryResult, rb_intern("new"), 2, terms, candidates);
}

static VALUE
text_index_search(VALUE self, VALUE text)
{
    text_index_data_t *data;
    rast_error_t *error;
    rast_query_result_t *result;
    VALUE vpool;
    apr_pool_t *pool;

    vpool = pool_new();
    pool = get_pool(vpool);
    Data_Get_Struct(self, text_index_data_t, data);
    error = rast_text_index_search(data->index, StringValuePtr(text), 1,
                                   &result, pool);
    if (error != RAST_OK) {
        raise_error(error);
    }
    return query_result_new(result);
}

static char *
hash_get_string(apr_pool_t *pool, VALUE hash, const char *name)
{
    VALUE val;

    val = rb_hash_aref(hash, rb_str_new2(name));
    SafeStringValue(val);
    return apr_pstrdup(pool, RSTRING(val)->ptr);
}

static int
hash_get_bool(VALUE hash, const char *name)
{
    VALUE val;

    val = rb_hash_aref(hash, rb_str_new2(name));
    return RTEST(val);
}

static rast_type_e
hash_get_property_type(VALUE hash, const char *name)
{
    VALUE val;


    val = rb_hash_aref(hash, rb_str_new2(name));
    return NUM2INT(val);
}

static void
get_bool_option(VALUE options, const char *name, int *dst)
{
    VALUE value = rb_hash_aref(options, rb_str_new2(name));
    if (!NIL_P(value)) {
        *dst = RTEST(value);
    }
}

static void
get_int_option(VALUE options, const char *name, int *dst)
{
    VALUE value = rb_hash_aref(options, rb_str_new2(name));
    if (RTEST(value)) {
        *dst = NUM2INT(value);
    }
}

static void
get_string_option(VALUE options, const char *name, const char **dst)
{
    VALUE value = rb_hash_aref(options, rb_str_new2(name));
    if (RTEST(value)) {
        SafeStringValue(value);
        *dst = RSTRING(value)->ptr;
    }
}

static rast_property_t *
get_properties(apr_pool_t *pool,
               VALUE vproperties, rast_size_t *num_properties)
{
    rast_property_t *properties;
    int i;

    Check_Type(vproperties, T_ARRAY);
    properties = (rast_property_t *) apr_palloc(pool,
                                                sizeof(rast_property_t) *
                                                RARRAY(vproperties)->len);
    for (i = 0; i < RARRAY(vproperties)->len; i++) {
        VALUE vproperty = RARRAY(vproperties)->ptr[i];
        rast_property_t *property = properties + i;

        Check_Type(vproperty, T_HASH);
        property->name = hash_get_string(pool, vproperty, "name");
        property->type = hash_get_property_type(vproperty, "type");
        property->flags = 0;
        if (hash_get_bool(vproperty, "search")) {
            property->flags |= RAST_PROPERTY_FLAG_SEARCH;
        }
        if (hash_get_bool(vproperty, "text_search")) {
            property->flags |= RAST_PROPERTY_FLAG_TEXT_SEARCH;
        }
        if (hash_get_bool(vproperty, "full_text_search")) {
            property->flags |= RAST_PROPERTY_FLAG_FULL_TEXT_SEARCH;
        }
        if (hash_get_bool(vproperty, "unique")) {
            property->flags |= RAST_PROPERTY_FLAG_UNIQUE;
        }
    }
    *num_properties = RARRAY(vproperties)->len;
    return properties;
}

static VALUE
db_s_create(VALUE self, VALUE vname, VALUE voptions)
{
    VALUE vpool;
    apr_pool_t *pool;
    VALUE properties;
    rast_error_t *error;
    char *name;
    rast_db_create_option_t *options;

    vpool = pool_new();
    pool = get_pool(vpool);
    SafeStringValue(vname);
    name = RSTRING(vname)->ptr;
    options = rast_db_create_option_create(pool);
    Check_Type(voptions, T_HASH);
    get_int_option(voptions, "byte_order", (int *) &options->byte_order);
    get_int_option(voptions, "pos_block_size", &options->pos_block_size);
    get_string_option(voptions, "encoding", &options->encoding);
    get_bool_option(voptions, "preserve_text", &options->preserve_text);
    properties = rb_hash_aref(voptions, rb_str_new2("properties"));
    options->properties =
        get_properties(pool, properties, &options->num_properties);
    error = rast_db_create(name, options, pool);
    if (error != RAST_OK) {
        raise_error(error);
    }
    return Qnil;
}

typedef struct {
    rast_db_t *db;
    apr_pool_t *pool;
    int closed;
} db_data_t;

static void
db_free(db_data_t *data)
{
    if (data) {
        if (!data->closed) {
            rast_db_close(data->db);
        }
        apr_pool_destroy(data->pool);
        xfree(data);
    }
}

static VALUE
db_alloc(VALUE klass)
{
    return Data_Wrap_Struct(klass, NULL, db_free, NULL);
}

static VALUE
db_initialize(int argc, VALUE *argv, VALUE self)
{
    VALUE name, vflags, voptions, vpool;
    rast_error_t *error;
    apr_pool_t *pool, *tmp_pool;
    db_data_t *data;
    rast_db_t *db;
    int flags = RAST_RDWR;
    rast_db_open_option_t *options = NULL;

    rb_scan_args(argc, argv, "12", &name, &vflags, &voptions);
    if (!NIL_P(vflags)) {
        flags = NUM2INT(vflags);
    }
    if (!NIL_P(voptions)) {
        vpool = pool_new();
        tmp_pool = get_pool(vpool);
        options = rast_db_open_option_create(tmp_pool);
        get_int_option(voptions, "sync_threshold_chars",
                       &options->sync_threshold_chars);
    }
    apr_pool_create_ex(&pool, NULL, raise_on_pool_failure, NULL);
    SafeStringValue(name);
    error = rast_db_open(&db, RSTRING(name)->ptr, flags, options, pool);
    if (error != RAST_OK) {
        apr_pool_destroy(pool);
        raise_error(error);
    }
    data = ALLOC(db_data_t);
    data->db = db;
    data->pool = pool;
    data->closed = 0;
    DATA_PTR(self) = data;
    return Qnil;
}

static rast_db_t *
get_db(VALUE self)
{
    db_data_t *data;

    if (TYPE(self) != T_DATA ||
        RDATA(self)->dfree != (RUBY_DATA_FUNC) db_free) {
        rb_raise(rb_eTypeError,
                 "wrong argument type %s (expected Rast::DB)",
                 rb_obj_classname(self));
    }

    data = DATA_PTR(self);
    if (data->closed) {
        rb_raise(eError, "already closed db");
    }
    return data->db;
}

static VALUE
db_close(VALUE self)
{
    db_data_t *data;
    rast_error_t *error;

    Data_Get_Struct(self, db_data_t, data);
    error = rast_db_close(data->db);
    if (error != RAST_OK) {
        raise_error(error);
    }
    data->closed = 1;
    return Qnil;
}

static VALUE
db_s_open(int argc, VALUE *argv, VALUE self)
{
    VALUE db = rb_class_new_instance(argc, argv, self);
    if (rb_block_given_p()) {
        return rb_ensure(rb_yield, db, db_close, db);
    }

    return db;
}

static VALUE
db_byte_order(VALUE self)
{
    rast_db_t *db;

    db = get_db(self);
    return INT2NUM(rast_db_byte_order(db));
}

static VALUE
db_encoding(VALUE self)
{
    rast_db_t *db;

    db = get_db(self);
    return rb_tainted_str_new2(rast_db_encoding(db));
}

static VALUE
db_properties(VALUE self)
{
    rast_db_t *db;
    const rast_property_t *properties;
    rast_size_t num_properties;
    VALUE result;
    int i;

    db = get_db(self);
    properties = rast_db_properties(db, &num_properties);
    result = rb_ary_new();
    for (i = 0; i < num_properties; i++) {
        const rast_property_t *property = properties + i;
        VALUE val, search, text_search, full_text_search, unique;

        search = property->flags & RAST_PROPERTY_FLAG_SEARCH ? Qtrue : Qfalse;
        text_search =
            property->flags & RAST_PROPERTY_FLAG_TEXT_SEARCH ? Qtrue : Qfalse;
        full_text_search =
            property->flags & RAST_PROPERTY_FLAG_FULL_TEXT_SEARCH ?
            Qtrue : Qfalse;
        unique = property->flags & RAST_PROPERTY_FLAG_UNIQUE ? Qtrue : Qfalse;
        val = rb_funcall(cProperty, rb_intern("new"), 6,
                         rb_tainted_str_new2(property->name),
                         INT2NUM(property->type),
                         search, text_search, full_text_search, unique);
        rb_ary_push(result, val);
    }
    return result;
}

static VALUE
db_sync_threshold_chars(VALUE self)
{
    rast_db_t *db;

    db = get_db(self);
    return INT2NUM(rast_db_sync_threshold_chars(db));
}

static rast_value_t *
get_property_values(VALUE vproperty_values, rast_db_t *db, apr_pool_t *pool)
{
    const rast_property_t *properties;
    rast_value_t *property_values;
    rast_size_t num_properties;
    int i;

    properties = rast_db_properties(db, &num_properties);
    property_values = (rast_value_t *)
        apr_palloc(pool, sizeof(rast_value_t) * num_properties);
    for (i = 0; i < num_properties; i++) {
        const rast_property_t *property = properties + i;
        VALUE value;

        value = rb_hash_aref(vproperty_values, rb_str_new2(property->name));
        switch (property->type) {
        case RAST_TYPE_STRING:
        case RAST_TYPE_DATE:
            Check_Type(value, T_STRING);
            rast_value_set_string(property_values + i, StringValuePtr(value));
            break;
        case RAST_TYPE_UINT:
            Check_Type(value, T_FIXNUM);
            rast_value_set_uint(property_values + i, NUM2INT(value));
            break;
        default:
            rb_raise(eError, "unknown property type");
        };
    }
    return property_values;
}

static VALUE
db_register(VALUE self, VALUE text, VALUE vproperty_values)
{
    rast_db_t *db;
    rast_value_t *property_values;
    rast_error_t *error;
    apr_pool_t *pool;
    VALUE vpool = pool_new();
    rast_doc_id_t doc_id;
    
    pool = get_pool(vpool);
    db = get_db(self);
    property_values = get_property_values(vproperty_values, db, pool);
    StringValue(text);
    error = rast_db_register(db, RSTRING(text)->ptr, RSTRING(text)->len,
                             property_values, &doc_id);
    if (error != RAST_OK) {
        raise_error(error);
    }
    return INT2NUM(doc_id);
}

static VALUE
result_new(rast_result_t *result, int num_properties)
{
    VALUE terms, term, items, vitem, properties;
    int i, j;

    terms = rb_ary_new();
    for (i = 0; i < result->num_terms; i++) {
        term = rb_funcall(cResultTerm, rb_intern("new"), 2,
                          rb_tainted_str_new2(result->terms[i].term),
                          INT2NUM(result->terms[i].doc_count));
        rb_ary_push(terms, term);
    }

    items = rb_ary_new();
    for (i = 0; i < result->num_items; i++) {
        rast_result_item_t *item = result->items[i];

        properties = rb_ary_new();
        for (j = 0; j < num_properties; j++) {
            rast_value_t *value = item->properties + j;
            switch (value->type) {
            case RAST_TYPE_STRING:
            case RAST_TYPE_DATE:
                rb_ary_push(properties,
                            rb_tainted_str_new2(rast_value_string(value)));
                break;
            case RAST_TYPE_UINT:
                rb_ary_push(properties, INT2NUM(rast_value_uint(value)));
                break;
            }
        }
        vitem = rb_funcall(cResultItem, rb_intern("new"), 4,
                           INT2NUM(item->doc_id),
                           INT2NUM(item->score),
                           rb_tainted_str_new(item->summary,
                                              item->summary_nbytes),
                           properties);
        rb_ary_push(items, vitem);
    }
    return rb_funcall(cResult, rb_intern("new"), 3,
                      INT2NUM(result->hit_count), terms, items);
}

static VALUE
db_sync(VALUE self)
{
    rast_db_t *db;
    rast_error_t *error;
    
    db = get_db(self);
    error = rast_db_sync(db);
    if (error != RAST_OK) {
        raise_error(error);
    }
    return Qnil;
}

static VALUE
db_search(int argc, VALUE *argv, VALUE self)
{
    VALUE query, voptions, properties;
    rast_db_t *db;
    rast_error_t *error;
    rast_search_option_t *options;
    rast_result_t *result;
    int i;
    apr_pool_t *pool;
    VALUE vpool = pool_new(), value;
    
    pool = get_pool(vpool);
    db = get_db(self);
    options = rast_search_option_create(pool);
    if (rb_scan_args(argc, argv, "11", &query, &voptions) == 2) {
        Check_Type(voptions, T_HASH);
        
        get_int_option(voptions, "start_no", &options->start_no);
        get_int_option(voptions, "num_items", &options->num_items);
        options->need_summary = hash_get_bool(voptions, "need_summary");
        get_int_option(voptions, "summary_nchars", &options->summary_nchars);
        get_int_option(voptions, "sort_order", (int *) &options->sort_order);
        value = rb_hash_aref(voptions, rb_str_new2("sort_property"));
        if (!NIL_P(value)) {
            options->sort_property = apr_pstrdup(pool, StringValuePtr(value));
        }
        get_int_option(voptions, "sort_method", (int *) &options->sort_method);
        get_int_option(voptions, "score_method",
                       (int *) &options->score_method);
        properties = rb_hash_aref(voptions, rb_str_new2("properties"));
        if (!NIL_P(properties)) {
            int properties_len = sizeof(char *) * RARRAY(properties)->len;

            Check_Type(properties, T_ARRAY);
            options->properties =
                (const char **) apr_palloc(pool, properties_len);
            for (i = 0; i < RARRAY(properties)->len; i++) {
                options->properties[i] =
                    StringValuePtr(RARRAY(properties)->ptr[i]);
            }
            options->num_properties = RARRAY(properties)->len;
        }
    }
    error = rast_db_search(db, StringValuePtr(query), options, &result, pool);
    if (error != RAST_OK) {
        raise_error(error);
    }
    return result_new(result, options->num_properties);
}

static VALUE
db_delete(VALUE self, VALUE vdoc_id)
{
    rast_doc_id_t doc_id;
    rast_db_t *db;
    rast_error_t *error;

    doc_id = (rast_doc_id_t) NUM2INT(vdoc_id);
    db = get_db(self);
    error = rast_db_delete(db, doc_id);
    if (error != RAST_OK) {
        raise_error(error);
    }
    return Qnil;
}

static VALUE
db_update(VALUE self, VALUE vdoc_id, VALUE text, VALUE vproperty_values)
{
    rast_db_t *db;
    rast_doc_id_t doc_id, new_doc_id;
    rast_value_t *property_values;
    rast_error_t *error;
    apr_pool_t *pool;
    VALUE vpool = pool_new();
    
    doc_id = (rast_doc_id_t) NUM2INT(vdoc_id);
    pool = get_pool(vpool);
    db = get_db(self);
    property_values = get_property_values(vproperty_values, db, pool);
    StringValue(text);
    error = rast_db_update(db, doc_id, RSTRING(text)->ptr, RSTRING(text)->len,
                           property_values, &new_doc_id);
    if (error != RAST_OK) {
        raise_error(error);
    }
    return INT2NUM(new_doc_id);
}

static VALUE
db_get_text(VALUE self, VALUE doc_id)
{
    rast_db_t *db;
    rast_error_t *error;
    apr_pool_t *pool;
    VALUE vpool = pool_new();
    char *s;
    rast_size_t len;

    pool = get_pool(vpool);
    db = get_db(self);
    error = rast_db_get_text(db, NUM2INT(doc_id), &s, &len, pool);
    if (error != RAST_OK) {
        raise_error(error);
    }
    return rb_tainted_str_new(s, len);
}

typedef struct {
    rast_query_t *query;
    apr_pool_t *pool;
} query_data_t;

static void
query_free(query_data_t *data)
{
    if (data) {
        apr_pool_destroy(data->pool);
        xfree(data);
    }
}

static VALUE
query_alloc(VALUE klass)
{
    return Data_Wrap_Struct(klass, NULL, query_free, NULL);
}

static VALUE
query_initialize(VALUE self, VALUE str)
{
    rast_error_t *error;
    apr_pool_t *pool;
    rast_query_t *query;
    query_data_t *data;
    rast_encoding_module_t *encoding_module;

    SafeStringValue(str);
    apr_pool_create_ex(&pool, NULL, raise_on_pool_failure, NULL);
    encoding_module = get_encoding_module("utf8");
    error = rast_parse_query(pool, encoding_module,
                             RSTRING(str)->ptr, &query);
    if (error != RAST_OK) {
        apr_pool_destroy(pool);
        raise_error(error);
    }
    data = ALLOC(query_data_t);
    data->query = query;
    data->pool = pool;
    DATA_PTR(self) = data;
    return Qnil;
}

static VALUE
query_inspect(VALUE self)
{
    query_data_t *data;
    VALUE vpool;
    apr_pool_t *pool;
    char *s;

    Data_Get_Struct(self, query_data_t, data);
    vpool = pool_new();
    pool = get_pool(vpool);
    s = rast_query_inspect(data->query, pool);
    return rb_tainted_str_new2(s);
}

static VALUE
query_exec(int argc, VALUE *argv, VALUE self)
{
    VALUE vdb, voptions;
    rast_error_t *error;
    query_data_t *data;
    rast_db_t *db;
    rast_query_result_t *result;
    VALUE vpool;
    apr_pool_t *pool;
    rast_query_option_t options;

    rb_scan_args(argc, argv, "11", &vdb, &voptions);
    options.score_method = RAST_SCORE_METHOD_TFIDF;
    if (!NIL_P(voptions)) {
        get_int_option(voptions, "score_method",
                       (int *) &options.score_method);
    }
    Data_Get_Struct(self, query_data_t, data);
    db = get_db(vdb);
    vpool = pool_new();
    pool = get_pool(vpool);
    error = rast_query_exec(data->query, db, &options, &result, pool);
    if (error != RAST_OK) {
        raise_error(error);
    }
    return query_result_new(result);
}

static VALUE
query_optimize(VALUE self)
{
    rast_error_t *error;
    query_data_t *data;

    Data_Get_Struct(self, query_data_t, data);
    error = rast_query_optimize(data->query, &data->query, data->pool);
    if (error != RAST_OK) {
        raise_error(error);
    }
    return self;
}

static int pack_number(char *s, int n)
{
    char *p = s;
    div_t d;

    if (n == 0) {
        *p = 0;
        return 1;
    }
    while (n > 0) {
        d = div(n, 128);
        n = d.quot;
        if (n > 0) {
            *p = d.rem | 0x80;
        }
        else {
            *p = d.rem;
        }
        p++;
    }
    return p - s;
}

static int unpack_number(const char *s, int nbytes, int *np)
{
    const char *p = s, *pend = s + nbytes;
    int base = 1, n = 0;

    while (p < pend) {
        if (*p & 0x80) {
            n += base * (*p & 0x7F);
            base *= 128;
            p++;
        }
        else {
            n += *p * base;
            p++;
            break;
        }
    }
    *np = n;
    return p - s;
}

static VALUE vnum_pack(VALUE self, VALUE ary)
{
    int i;
    VALUE result;

    Check_Type(ary, T_ARRAY);
    result = rb_str_new("", 0);
    for (i = 0; i < RARRAY(ary)->len; i++) {
        char buf[5];
        int num, nbytes;
        num = NUM2INT(RARRAY(ary)->ptr[i]);
        if (num < 0) {
            rb_raise(rb_eArgError, "negative value - %d", num);
        }
        nbytes = pack_number(buf, num);
        rb_str_cat(result, buf, nbytes);
    }
    return result;
}

static VALUE vnum_unpack(VALUE self, VALUE str)
{
    VALUE result;
    char *p, *pend;

    Check_Type(str, T_STRING);
    p = RSTRING(str)->ptr;
    pend = p + RSTRING(str)->len;
    result = rb_ary_new();
    while (p < pend) {
        int num, nbytes;
        nbytes = unpack_number(p, pend - p, &num);
        p += nbytes;
        rb_ary_push(result, INT2NUM(num));
    }
    return result;
}

void
Init_rast()
{
    VALUE mRast, cTextIndex, cDB, cQuery, utf8, euc_jp;

    apr_initialize();
    atexit(apr_terminate);
    rast_initialize();

    mRast = rb_define_module("Rast");

    eError = rb_define_class_under(mRast, "Error", rb_eStandardError);
    eRastError = rb_define_class_under(mRast, "RastError", eError);
    eAprError = rb_define_class_under(mRast, "AprError", eError);
    eBDBError = rb_define_class_under(mRast, "BDBError", eError);

    mVNUM = rb_define_module_under(mRast, "VNUM");
    rb_define_module_function(mVNUM, "pack", vnum_pack, 1);
    rb_define_module_function(mVNUM, "unpack", vnum_unpack, 1);

    rb_define_const(mRast, "VERSION", rb_str_new2(VERSION));

    rb_define_const(mRast, "NATIVE_ENDIAN", INT2NUM(RAST_NATIVE_ENDIAN));
    rb_define_const(mRast, "LITTLE_ENDIAN", INT2NUM(RAST_LITTLE_ENDIAN));
    rb_define_const(mRast, "BIG_ENDIAN", INT2NUM(RAST_BIG_ENDIAN));

    cPool = rb_define_class_under(mRast, "Pool", rb_cObject);
    rb_define_alloc_func(cPool, pool_alloc);
    rb_define_method(cPool, "initialize", pool_initialize, 0);

    cEncoding = rb_define_class_under(mRast, "Encoding", rb_cObject);
    rb_undef_method(CLASS_OF(cEncoding), "new");
    rb_define_method(cEncoding, "register_tokenize",
                     encoding_register_tokenize, 1);
    rb_define_method(cEncoding, "search_tokenize",
                     encoding_search_tokenize, 1);
    rb_define_method(cEncoding, "normalize_text",
                     encoding_normalize_text, 1);
    rb_define_method(cEncoding, "normalize_chars",
                     encoding_normalize_chars, 1);
    utf8 = encoding_new("utf8");
    rb_define_const(cEncoding, "UTF8", utf8);
    euc_jp = encoding_new("euc_jp");
    rb_define_const(cEncoding, "EucJp", euc_jp);

    cTerm = rb_funcall(rb_cStruct, rb_intern("new"), 2,
                       ID2SYM(rb_intern("term")),
                       ID2SYM(rb_intern("doc_count")));
    rb_define_const(mRast, "Term", cTerm);
    cTermFrequency = rb_funcall(rb_cStruct, rb_intern("new"), 2,
                                ID2SYM(rb_intern("count")),
                                ID2SYM(rb_intern("pos")));
    rb_define_const(mRast, "TermFrequency", cTermFrequency);
    cCandidate = rb_funcall(rb_cStruct, rb_intern("new"), 2,
                              ID2SYM(rb_intern("doc_id")),
                              ID2SYM(rb_intern("terms")));
    rb_define_const(mRast, "Candidate", cCandidate);
    cQueryResult = rb_funcall(rb_cStruct, rb_intern("new"), 2,
                              ID2SYM(rb_intern("terms")),
                              ID2SYM(rb_intern("candidates")));
    rb_define_const(mRast, "QueryResult", cQueryResult);
    cResultTerm = rb_funcall(rb_cStruct, rb_intern("new"), 2,
                             ID2SYM(rb_intern("term")),
                             ID2SYM(rb_intern("doc_count")));
    rb_define_const(mRast, "ResultTerm", cResultTerm);
    cResultItem = rb_funcall(rb_cStruct, rb_intern("new"), 4,
                             ID2SYM(rb_intern("doc_id")),
                             ID2SYM(rb_intern("score")),
                             ID2SYM(rb_intern("summary")),
                             ID2SYM(rb_intern("properties")));
    rb_define_const(mRast, "ResultItem", cResultItem);
    cResult = rb_funcall(rb_cStruct, rb_intern("new"), 3,
                         ID2SYM(rb_intern("hit_count")),
                         ID2SYM(rb_intern("terms")),
                         ID2SYM(rb_intern("items")));
    rb_define_const(mRast, "Result", cResult);

    cTextIndexer = rb_define_class_under(mRast, "TextIndexer", rb_cObject);
    rb_undef_method(CLASS_OF(cTextIndexer), "new");
    rb_define_method(cTextIndexer, "add", text_indexer_add, 1);
    rb_define_method(cTextIndexer, "commit", text_indexer_commit, 0);

    cTextIndex = rb_define_class_under(mRast, "TextIndex", rb_cObject);
    rb_define_alloc_func(cTextIndex, text_index_alloc);
    rb_define_method(cTextIndex, "initialize", text_index_initialize, -1);
    rb_define_method(cTextIndex, "close", text_index_close, 0);
    rb_define_method(cTextIndex, "sync", text_index_sync, 0);
    rb_define_method(cTextIndex, "register", text_index_register, -1);
    rb_define_method(cTextIndex, "search", text_index_search, 1);

    cProperty = rb_funcall(rb_cStruct, rb_intern("new"), 6,
                           ID2SYM(rb_intern("name")),
                           ID2SYM(rb_intern("type")),
                           ID2SYM(rb_intern("search")),
                           ID2SYM(rb_intern("text_search")),
                           ID2SYM(rb_intern("full_text_search")),
                           ID2SYM(rb_intern("unique")));
    rb_define_const(mRast, "Property", cProperty);

    rb_define_const(mRast, "PROPERTY_TYPE_STRING",
                    INT2NUM(RAST_TYPE_STRING));
    rb_define_const(mRast, "PROPERTY_TYPE_UINT",
                    INT2NUM(RAST_TYPE_UINT));
    rb_define_const(mRast, "PROPERTY_TYPE_DATE",
                    INT2NUM(RAST_TYPE_DATE));

    rb_define_const(mRast, "SORT_METHOD_SCORE",
                    INT2NUM(RAST_SORT_METHOD_SCORE));
    rb_define_const(mRast, "SORT_METHOD_PROPERTY",
                    INT2NUM(RAST_SORT_METHOD_PROPERTY));

    rb_define_const(mRast, "SORT_ORDER_DEFAULT",
                    INT2NUM(RAST_SORT_ORDER_DEFAULT));
    rb_define_const(mRast, "SORT_ORDER_ASCENDING",
                    INT2NUM(RAST_SORT_ORDER_ASCENDING));
    rb_define_const(mRast, "SORT_ORDER_DESCENDING",
                    INT2NUM(RAST_SORT_ORDER_DESCENDING));

    rb_define_const(mRast, "SCORE_METHOD_NONE",
                    INT2NUM(RAST_SCORE_METHOD_NONE));
    rb_define_const(mRast, "SCORE_METHOD_TFIDF",
                    INT2NUM(RAST_SCORE_METHOD_TFIDF));

    cDB = rb_define_class_under(mRast, "DB", rb_cObject);
    rb_define_const(cDB, "RDWR", INT2NUM(RAST_RDWR));
    rb_define_const(cDB, "RDONLY", INT2NUM(RAST_RDONLY));
    rb_define_singleton_method(cDB, "create", db_s_create, 2);
    rb_define_singleton_method(cDB, "open", db_s_open, -1);
    rb_define_alloc_func(cDB, db_alloc);
    rb_define_method(cDB, "initialize", db_initialize, -1);
    rb_define_method(cDB, "byte_order", db_byte_order, 0);
    rb_define_method(cDB, "encoding", db_encoding, 0);
    rb_define_method(cDB, "properties", db_properties, 0);
    rb_define_method(cDB, "sync_threshold_chars", db_sync_threshold_chars, 0);
    rb_define_method(cDB, "register", db_register, 2);
    rb_define_method(cDB, "sync", db_sync, 0);
    rb_define_method(cDB, "search", db_search, -1);
    rb_define_method(cDB, "delete", db_delete, 1);
    rb_define_method(cDB, "update", db_update, 3);
    rb_define_method(cDB, "get_text", db_get_text, 1);
    rb_define_method(cDB, "close", db_close, 0);

    cQuery = rb_define_class_under(mRast, "Query", rb_cObject);
    rb_define_alias(CLASS_OF(cQuery), "parse", "new");
    rb_define_alloc_func(cQuery, query_alloc);
    rb_define_method(cQuery, "initialize", query_initialize, 1);
    rb_define_method(cQuery, "inspect", query_inspect, 0);
    rb_define_method(cQuery, "exec", query_exec, -1);
    rb_define_method(cQuery, "optimize", query_optimize, 0);
}

/* vim: set filetype=c sw=4 expandtab : */
